#define XMOUSE_VER_ADDR (0xf000)
#define XMOUSE_HANDLER1 (0xf100)

#define MASK_MOTION_EVENT (1)
#define MASK_LEFT_PRESS_EVENT (2)
#define MASK_LEFT_RELEASE_EVENT (4)
#define MASK_RIGHT_PRESS_EVENT (8)
#define MASK_RIGHT_RELEASE_EVENT (16)
#define MASK_MID_PRESS_EVENT (32)
#define MASK_MID_RELEASE_EVENT (64)
