/* config.h, DOS emulator configuration file, Robert Sanders,
 * gt8134b@prism.gatech.edu, generated by dosconfig.c
 * 
 * DO NOT ALTER! "make config" will erase your changes!
 */

#ifndef CONFIG_H
#define CONFIG_H 1

#define    CONFIG_HOST     "linux, Linux 0.99.14 #5 Sat Jan 29 04:00:50 CST 1994"
#define    CONFIG_TIME     "Thu Feb 10 16:06:37 1994"
#define    linux           1              
#define    KBD_US                         
#define    KBDFLAGS        0              
#define    VGA_VIDEO                      
#define    CONFIG_FILE     "/etc/dosemu/config"
#define    CHEAP_GFX       1              
#define    DEF_FDISKS      2              
#define    DEF_HDISKS      2              
#define    VERNUM          0x49           
#define    VERSTR          "0.49pl3"      
#define    XMS             1              

#endif /* config.h */

/* total defines:  14
 * end of configuration information
 */
