DOS_BOOT_DEV=a
export DOS_BOOT_DEV
#the no. of true HDs DOS will see, excluding emu file sys. 
DOS_HDS=0
export DOS_HDS
DOS_USE_FDIMAGE=fdimage
export DOS_USE_FDIMAGE
./xdos -g 80x25 -sb -sl 1000 -fn 12x24 -fnch tt -cr red -fg yellow -bg black -bd green \
-xms -mousedrv $*
